@@ -1,6 +1,6 @@
-/* $Id: DBIXS.h,v 11.2 2001/08/24 22:10:44 timbo Exp $
+/* $Id: DBIXS.h,v 11.4 2002/05/20 21:03:53 timbo Exp $
  *
- * Copyright (c) 1994, 1995, 1996, 1997, 1998, 1999  Tim Bunce  England
+ * Copyright (c) 1994-2002  Tim Bunce  Ireland
  *
  * See COPYRIGHT section in DBI.pm for usage and distribution rights.
  */
@@ -112,7 +112,7 @@ typedef struct dbih_com_attr_st {
     SV *Handlers;
     U32  LongReadLen;	/* auto read length for long/blob types	*/
     SV *FetchHashKeyName;	/* for fetchrow_hashref		*/
-    /* (new fields?... don't forget to update dbih_clearcom()!)	*/
+    /* (NEW FIELDS?... DON'T FORGET TO UPDATE dbih_clearcom()!)	*/
 } dbih_com_attr_t;
 
 
@@ -192,6 +192,7 @@ typedef struct {		/* -- FIELD DESCRIPTOR --		*/
 #define DBIc_IMP_DATA(imp)  	_imp2com(imp, std.imp_data)
 #define DBIc_KIDS(imp)  	_imp2com(imp, std.kids)
 #define DBIc_ACTIVE_KIDS(imp)  	_imp2com(imp, std.active_kids)
+#define DBIc_LAST_METHOD(imp)  	_imp2com(imp, std.last_method)
 
 #define DBIc_DEBUG(imp)		(_imp2com(imp, attr.Debug))
 #define DBIc_DEBUGIV(imp)	SvIV(DBIc_DEBUG(imp))
@@ -215,22 +216,23 @@ typedef struct {		/* -- FIELD DESCRIPTOR --		*/
 #define DBIc_FDESC_AV(imp)  	_imp2com(imp, fields_fdav)
 #define DBIc_FDESC(imp, i)  	((imp_fdh_t*)(void*)SvPVX(AvARRAY(DBIc_FDESC_AV(imp))[i]))
 
-#define DBIcf_COMSET	  0x0001	/* needs to be clear'd before free'd	*/
-#define DBIcf_IMPSET	  0x0002	/* has implementor data to be clear'd	*/
-#define DBIcf_ACTIVE	  0x0004	/* needs finish/disconnect before clear	*/
-#define DBIcf_IADESTROY	  0x0008	/* do DBIc_ACTIVE_off before DESTROY	*/
-#define DBIcf_WARN  	  0x0010	/* warn about poor practice etc  	*/
-#define DBIcf_COMPAT  	  0x0020	/* compat/emulation mode (eg oraperl)	*/
+#define DBIcf_COMSET	  0x000001	/* needs to be clear'd before free'd	*/
+#define DBIcf_IMPSET	  0x000002	/* has implementor data to be clear'd	*/
+#define DBIcf_ACTIVE	  0x000004	/* needs finish/disconnect before clear	*/
+#define DBIcf_IADESTROY	  0x000008	/* do DBIc_ACTIVE_off before DESTROY	*/
+#define DBIcf_WARN  	  0x000010	/* warn about poor practice etc  	*/
+#define DBIcf_COMPAT  	  0x000020	/* compat/emulation mode (eg oraperl)	*/
 
-#define DBIcf_ChopBlanks  0x0040	/* rtrim spaces from fetch char columns	*/
-#define DBIcf_RaiseError  0x0080	/* throw exception (croak) on error	*/
-#define DBIcf_PrintError  0x0100	/* warn() on error			*/
-#define DBIcf_AutoCommit  0x0200	/* dbh only. used by drivers		*/
-#define DBIcf_LongTruncOk 0x0400	/* truncation to LongReadLen is okay	*/
-#define DBIcf_MultiThread 0x0800	/* allow multiple threads to enter	*/
-#define DBIcf_Taint       0x1000	/* taint fetched data			*/
-#define DBIcf_ShowErrorStatement  0x2000	/* include Statement in error	*/
-#define DBIcf_BegunWork   0x4000	/* between begin_work & commit/rollback */
+#define DBIcf_ChopBlanks  0x000040	/* rtrim spaces from fetch char columns	*/
+#define DBIcf_RaiseError  0x000080	/* throw exception (croak) on error	*/
+#define DBIcf_PrintError  0x000100	/* warn() on error			*/
+#define DBIcf_AutoCommit  0x000200	/* dbh only. used by drivers		*/
+#define DBIcf_LongTruncOk 0x000400	/* truncation to LongReadLen is okay	*/
+#define DBIcf_MultiThread 0x000800	/* allow multiple threads to enter	*/
+#define DBIcf_Taint       0x001000	/* taint fetched data			*/
+#define DBIcf_ShowErrorStatement  0x002000	/* include Statement in error	*/
+#define DBIcf_BegunWork   0x004000	/* between begin_work & commit/rollback */
+#define DBIcf_HandleError 0x008000	/* has coderef in HandleError attribute	*/
 
 #define DBIcf_INHERITMASK		/* what NOT to pass on to children */	\
   (U32)( DBIcf_COMSET | DBIcf_IMPSET | DBIcf_ACTIVE | DBIcf_IADESTROY		\
@@ -381,7 +383,8 @@ typedef struct {
     dbi_mutex	*mutex;
 
     int         (*logmsg)	_((imp_xxh_t *imp_xxh, char *fmt, ...));
-    void *pad[8];
+    int         (*set_err)	_((imp_xxh_t *imp_xxh, char *fmt, ...));
+    void *pad[7];
 } dbistate_t;
 
 /* macros for backwards compatibility */
@@ -402,7 +405,7 @@ typedef struct {
 /* note that USE_ITHREADS implies MULTIPLICITY                      */
 #if defined(MULTIPLICITY) || defined(PERL_OBJECT) || defined(PERL_CAPI)
 
-# define DBISTATE_DECLARE
+# define DBISTATE_DECLARE typedef int dummy_dbistate /* keep semicolon from feeling lonely */
 # define DBISTATE_ASSIGN(st)
 # define DBISTATE_INIT
 static dbistate_t **get_dbistate() {
@@ -438,9 +441,10 @@ static dbistate_t **get_dbistate() {
 /* attribs value. One day we may add some extra magic in here.		*/
 #define DBD_ATTRIBS_CHECK(func, h, attribs)	\
     if ((attribs) && SvOK(attribs)) {		\
+	STRLEN lna=0;				\
 	if (!SvROK(attribs) || SvTYPE(SvRV(attribs))!=SVt_PVHV)		\
 	    croak("%s->%s(...): attribute parameter '%s' is not a hash ref",	\
-		    SvPV(h,na), func, SvPV(attribs,na));		\
+		    SvPV(h,lna), func, SvPV(attribs,lna));		\
     } else (attribs) = Nullsv
 
 #define DBD_ATTRIB_GET_SVP(attribs, key,klen)			\
