@@ -176,12 +176,15 @@ use vars qw(%ATTR $AUTOLOAD);
     'PrintError' => 'local',
     'RaiseError' => 'local',
     'RowCacheSize' => 'inherited',
-    'AutoCommit' => 'cached'
+    'AutoCommit' => 'cached',
+    'Database' => 'cached',
+    'Statement' => 'local',
 );
 
 sub AUTOLOAD {
     my $method = $AUTOLOAD;
     $method =~ s/(.*::(.*)):://;
+    # warn "AUTOLOAD of $method";
     my $class = $1;
     my $type = $2;
     my %expand =
@@ -307,6 +310,7 @@ sub quote {
 sub table_info {
     my $dbh = shift;
     my $rdbh = $dbh->{'proxy_dbh'};
+    #warn "table_info(@_)";
     my($numFields, $names, $types, @rows) = eval { $rdbh->table_info(@_) };
     return DBI::set_err($dbh, 1, $@) if $@;
     my $sth = DBI::_new_sth($dbh, {
@@ -325,6 +329,13 @@ sub table_info {
     return $sth;
 }
 
+sub tables {
+    my $dbh = shift;
+    #warn "tables(@_)";
+    return $dbh->SUPER::tables(@_);
+}
+
+
 sub type_info_all {
     my $dbh = shift;
     my $result = eval { $dbh->{'proxy_dbh'}->type_info_all(@_) };
@@ -518,7 +529,7 @@ sub FETCH ($$) {
     }
 
     if ($type eq 'cache_only'  &&
-	exists($sth->{'proxy_attr_cache'}->{$attr})) {
+	    exists($sth->{'proxy_attr_cache'}->{$attr})) {
 	return $sth->{'proxy_attr_cache'}->{$attr};
     }
 
@@ -527,10 +538,12 @@ sub FETCH ($$) {
 	my $result = eval { $rsth->FETCH($attr) };
 	return DBI::set_err($sth, 1, $@) if $@;
 	return $result;
-    } elsif ($attr eq 'RowsInCache') {
+    }
+    elsif ($attr eq 'RowsInCache') {
 	my $data = $sth->{'proxy_data'};
 	$data ? @$data : 0;
-    } else {
+    }
+    else {
 	$sth->SUPER::FETCH($attr);
     }
 }
