@@ -1,6 +1,6 @@
 #!/usr/local/bin/perl -w
 
-# $Id: test.pl,v 11.3 2002/01/10 15:14:06 timbo Exp $
+# $Id: test.pl,v 11.4 2002/05/20 21:03:53 timbo Exp $
 #
 # Copyright (c) 1994-1998 Tim Bunce
 #
@@ -14,7 +14,7 @@
 
 BEGIN {
     print "$0 @ARGV\n";
-    print q{DBI test application $Revision: 11.3 $}."\n";
+    print q{DBI test application $Revision: 11.4 $}."\n";
     $| = 1;
     eval "require blib; import blib;";	# wasn't in 5.003, hence the eval
     warn $@ if $@;
@@ -71,14 +71,15 @@ if (0) {
 
 if ($::opt_m) {
     #$dbh->trace(9);
+    my $level = 4;
     my $cnt = 10000;
     print "Using $driver, same dbh...\n";
-    #for (my $i=0; $i<$cnt; ++$i) { mem_test($dbh, undef, 4) }
+    #for (my $i=0; $i<$cnt; ++$i) { mem_test($dbh, undef, $level) }
     print "Using NullP, reconnecting each time...\n";
-    #for (my $i=0; $i<$cnt; ++$i) { mem_test(undef, ['dbi:NullP:'], 4) }
+    #for (my $i=0; $i<$cnt; ++$i) { mem_test(undef, ['dbi:NullP:'], $level) }
     print "Using ExampleP, reconnecting each time...\n";
-    mem_test(undef, ['dbi:ExampleP:'], 4) while 1;
-    #mem_test(undef, ['dbi:mysql:VC'], 4, "select * from campaigns where length(?)>0") while 1;
+    mem_test(undef, ['dbi:ExampleP:'], $level) while 1;
+    #mem_test(undef, ['dbi:mysql:VC'], $level, "select * from campaigns where length(?)>0") while 1;
 }
 elsif ($::opt_t) {
 	thread_test();
@@ -94,13 +95,13 @@ else {
     print "Testing handle creation speed...\n";
     my $null_dbh = DBI->connect('dbi:NullP:','','');
     my $null_sth = $null_dbh->prepare('');	# create one to warm up
-    $count = 5000;
+    $count = 10_000;
     my $i = $count;
     my $t1 = new Benchmark;
     $null_dbh->prepare('') while $i--;
     my $td = Benchmark::timediff(Benchmark->new, $t1);
     my $tds= Benchmark::timestr($td);
-    my $dur = $td->cpu_a;
+    my $dur = $td->cpu_a || (1/$count); # fudge if cpu_a==0
     printf "$count NullP statement handles cycled in %.1f cpu+sys seconds (%d per sec)\n\n",
 	    $dur, $count / $dur;
 
